# Copyright (C) 2013-2019 YouCompleteMe contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
from ycmd.utils import ToUnicode
from ycm.client.base_request import ( BaseRequest,
                                      DisplayServerException,
                                      MakeServerException )
from ycm import vimsupport, base
from ycm.vimsupport import NO_COMPLETIONS

_logger = logging.getLogger( __name__ )


class CompletionRequest( BaseRequest ):
  def __init__( self, request_data ):
    super().__init__()
    self.request_data = request_data
    self._response_future = None


  def Start( self ):
    self._response_future = self.PostDataToHandlerAsync( self.request_data,
                                                         'completions' )


  def Done( self ):
    return bool( self._response_future ) and self._response_future.done()


  def _RawResponse( self ):
    if not self._response_future:
      return NO_COMPLETIONS

    response = self.HandleFuture( self._response_future,
                                  truncate_message = True )
    if not response:
      return NO_COMPLETIONS

    # Vim may not be able to convert the 'errors' entry to its internal format
    # so we remove it from the response.
    errors = response.pop( 'errors', [] )
    for e in errors:
      exception = MakeServerException( e )
      _logger.error( exception )
      DisplayServerException( exception, truncate_message = True )

    response[ 'line' ] = self.request_data[ 'line_num' ]
    response[ 'column' ] = self.request_data[ 'column_num' ]
    return response


  def Response( self ):
    response = self._RawResponse()
    response[ 'completions' ] = _ConvertCompletionDatasToVimDatas(
        response[ 'completions' ] )
    # FIXME: Do we really need to do this AdjustCandidateInsertionText ? I feel
    # like Vim should do that for us
    response[ 'completions' ] = base.AdjustCandidateInsertionText(
        response[ 'completions' ] )
    return response


  def OnCompleteDone( self ):
    if not self.Done():
      return

    if 'cs' in vimsupport.CurrentFiletypes():
      self._OnCompleteDone_Csharp()
    else:
      self._OnCompleteDone_FixIt()


  def _GetExtraDataUserMayHaveCompleted( self ):
    completed_item = vimsupport.GetVariableValue( 'v:completed_item' )

    # If Vim supports user_data (8.0.1493 or later), we actually know the
    # _exact_ element that was selected, having put its extra_data in the
    # user_data field. Otherwise, we have to guess by matching the values in the
    # completed item and the list of completions. Sometimes this returns
    # multiple possibilities, which is essentially unresolvable.
    if 'user_data' not in completed_item:
      completions = self._RawResponse()[ 'completions' ]
      return _FilterToMatchingCompletions( completed_item, completions )

    if completed_item[ 'user_data' ]:
      return [ json.loads( completed_item[ 'user_data' ] ) ]

    return []


  def _OnCompleteDone_Csharp( self ):
    extra_datas = self._GetExtraDataUserMayHaveCompleted()
    namespaces = [ _GetRequiredNamespaceImport( c ) for c in extra_datas ]
    namespaces = [ n for n in namespaces if n ]
    if not namespaces:
      return

    if len( namespaces ) > 1:
      choices = [ f"{ i + 1 } { n }" for i, n in enumerate( namespaces ) ]
      choice = vimsupport.PresentDialog( "Insert which namespace:", choices )
      if choice < 0:
        return
      namespace = namespaces[ choice ]
    else:
      namespace = namespaces[ 0 ]

    vimsupport.InsertNamespace( namespace )


  def _OnCompleteDone_FixIt( self ):
    extra_datas = self._GetExtraDataUserMayHaveCompleted()
    fixit_completions = [ _GetFixItCompletion( c ) for c in extra_datas ]
    fixit_completions = [ f for f in fixit_completions if f ]
    if not fixit_completions:
      return

    # If we have user_data in completions (8.0.1493 or later), then we would
    # only ever return max. 1 completion here. However, if we had to guess, it
    # is possible that we matched multiple completion items (e.g. for overloads,
    # or similar classes in multiple packages). In any case, rather than
    # prompting the user and disturbing her workflow, we just apply the first
    # one. This might be wrong, but the solution is to use a (very) new version
    # of Vim which supports user_data on completion items
    fixit_completion = fixit_completions[ 0 ]

    for fixit in fixit_completion:
      vimsupport.ReplaceChunks( fixit[ 'chunks' ], silent=True )


def _GetRequiredNamespaceImport( extra_data ):
  return extra_data.get( 'required_namespace_import' )


def _GetFixItCompletion( extra_data ):
  return extra_data.get( 'fixits' )


def _FilterToMatchingCompletions( completed_item, completions ):
  """Filter to completions matching the item Vim said was completed"""
  match_keys = [ 'word', 'abbr', 'menu', 'info' ]
  matched_completions = []
  for completion in completions:
    item = ConvertCompletionDataToVimData( completion )

    def matcher( key ):
      return ( ToUnicode( completed_item.get( key, "" ) ) ==
               ToUnicode( item.get( key, "" ) ) )

    if all( matcher( i ) for i in match_keys ):
      matched_completions.append( completion.get( 'extra_data', {} ) )
  return matched_completions


def _GetCompletionInfoField( completion_data ):
  info = completion_data.get( 'detailed_info', '' )

  if 'extra_data' in completion_data:
    docstring = completion_data[ 'extra_data' ].get( 'doc_string', '' )
    if docstring:
      if info:
        info += '\n' + docstring
      else:
        info = docstring

  # This field may contain null characters e.g. \x00 in Python docstrings. Vim
  # cannot evaluate such characters so they are removed.
  return info.replace( '\x00', '' )


def ConvertCompletionDataToVimData( completion_data ):
  # See :h complete-items for a description of the dictionary fields.
  extra_menu_info = completion_data.get( 'extra_menu_info', '' )
  preview_info = _GetCompletionInfoField( completion_data )

  # When we are using a popup for the preview_info, it needs to fit on the
  # screen alongside the extra_menu_info. Let's use some heuristics.  If the
  # length of the extra_menu_info is more than, say, 1/3 of screen, truncate it
  # and stick it in the preview_info.
  if vimsupport.UsingPreviewPopup():
    max_width = max( int( vimsupport.DisplayWidth() / 3 ), 3 )
    extra_menu_info_width = vimsupport.DisplayWidthOfString( extra_menu_info )
    if extra_menu_info_width > max_width:
      if not preview_info.startswith( extra_menu_info ):
        preview_info = extra_menu_info + '\n\n' + preview_info
      extra_menu_info = extra_menu_info[ : ( max_width - 3 ) ] + '...'

  return {
    'word'     : completion_data[ 'insertion_text' ],
    'abbr'     : completion_data.get( 'menu_text', '' ),
    'menu'     : extra_menu_info,
    'info'     : preview_info,
    'kind'     : ToUnicode( completion_data.get( 'kind', '' ) )[ :1 ].lower(),
    # Disable Vim filtering.
    'equal'    : 1,
    'dup'      : 1,
    'empty'    : 1,
    # We store the completion item extra_data as a string in the completion
    # user_data. This allows us to identify the _exact_ item that was completed
    # in the CompleteDone handler, by inspecting this item from v:completed_item
    #
    # We convert to string because completion user data items must be strings.
    #
    # Note: Not all versions of Vim support this (added in 8.0.1483), but adding
    # the item to the dictionary is harmless in earlier Vims.
    # Note: Since 8.2.0084 we don't need to use json.dumps() here.
    'user_data': json.dumps( completion_data.get( 'extra_data', {} ) )
  }


def _ConvertCompletionDatasToVimDatas( response_data ):
  return [ ConvertCompletionDataToVimData( x ) for x in response_data ]



# Copyright (C) 2013  Google Inc.
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

from ycm.client.base_request import ( BaseRequest,
                                      BuildRequestData,
                                      BuildRequestDataForLocation )
from ycm import vimsupport

DEFAULT_BUFFER_COMMAND = 'same-buffer'


def _EnsureBackwardsCompatibility( arguments ):
  if arguments and arguments[ 0 ] == 'GoToDefinitionElseDeclaration':
    arguments[ 0 ] = 'GoTo'
  return arguments


class CommandRequest( BaseRequest ):
  def __init__( self,
                arguments,
                extra_data = None,
                silent = False,
                location = None ):
    super( CommandRequest, self ).__init__()
    self._arguments = _EnsureBackwardsCompatibility( arguments )
    self._command = arguments and arguments[ 0 ]
    self._extra_data = extra_data
    self._response = None
    self._request_data = None
    self._response_future = None
    self._silent = silent
    self._bufnr = extra_data.pop( 'bufnr', None ) if extra_data else None
    self._location = location


  def Start( self ):
    if self._location is not None:
      self._request_data = BuildRequestDataForLocation( *self._location )
    elif self._bufnr is not None:
      self._request_data = BuildRequestData( self._bufnr )
    else:
      self._request_data = BuildRequestData()

    if self._extra_data:
      self._request_data.update( self._extra_data )
    self._request_data.update( {
      'command_arguments': self._arguments
    } )
    self._response_future = self.PostDataToHandlerAsync(
      self._request_data,
      'run_completer_command' )


  def Done( self ):
    return bool( self._response_future ) and self._response_future.done()


  def Response( self ):
    if self._response is None and self._response_future is not None:
      # Block
      self._response = self.HandleFuture( self._response_future,
                                          display_message = not self._silent )

    return self._response


  def RunPostCommandActionsIfNeeded( self,
                                     modifiers,
                                     buffer_command = DEFAULT_BUFFER_COMMAND ):

    # This is a blocking call if not Done()
    self.Response()

    if self._response is None:
      # An exception was raised and handled.
      return

    # If not a dictionary or a list, the response is necessarily a
    # scalar: boolean, number, string, etc. In this case, we print
    # it to the user.
    if not isinstance( self._response, ( dict, list ) ):
      return self._HandleBasicResponse()

    if 'fixits' in self._response:
      return self._HandleFixitResponse()

    if 'message' in self._response:
      return self._HandleMessageResponse()

    if 'detailed_info' in self._response:
      return self._HandleDetailedInfoResponse( modifiers )

    # The only other type of response we understand is GoTo, and that is the
    # only one that we can't detect just by inspecting the response (it should
    # either be a single location or a list)
    return self._HandleGotoResponse( buffer_command, modifiers )


  def StringResponse( self ):
    # Retuns a supporable public API version of the response. The reason this
    # exists is that the ycmd API here is wonky as it originally only supported
    # text-responses and now has things like fixits and such.
    #
    # The supportable public API is basically any text-only response. All other
    # response types are returned as empty strings

    # This is a blocking call if not Done()
    self.Response()

    # Completer threw an error ?
    if self._response is None:
      return ""

    # If not a dictionary or a list, the response is necessarily a
    # scalar: boolean, number, string, etc. In this case, we print
    # it to the user.
    if not isinstance( self._response, ( dict, list ) ):
      return str( self._response )

    if 'message' in self._response:
      return self._response[ 'message' ]

    if 'detailed_info' in self._response:
      return self._response[ 'detailed_info' ]

    # The only other type of response we understand is 'fixits' and GoTo. We
    # don't provide string versions of them.
    return ""


  def _HandleGotoResponse( self, buffer_command, modifiers ):
    if isinstance( self._response, list ):
      vimsupport.SetQuickFixList(
        [ vimsupport.BuildQfListItem( x ) for x in self._response ] )
      vimsupport.OpenQuickFixList( focus = True, autoclose = True )
    elif self._response.get( 'file_only' ):
      vimsupport.JumpToLocation( self._response[ 'filepath' ],
                                 None,
                                 None,
                                 modifiers,
                                 buffer_command )
    else:
      vimsupport.JumpToLocation( self._response[ 'filepath' ],
                                 self._response[ 'line_num' ],
                                 self._response[ 'column_num' ],
                                 modifiers,
                                 buffer_command )


  def _HandleFixitResponse( self ):
    if not len( self._response[ 'fixits' ] ):
      vimsupport.PostVimMessage( 'No fixits found for current line',
                                 warning = False )
    else:
      try:
        fixit_index = 0

        # If there is more than one fixit, we need to ask the user which one
        # should be applied.
        #
        # If there's only one, triggered by the FixIt subcommand (as opposed to
        # `RefactorRename`, for example) and whose `kind` is not `quicfix`, we
        # still need to as the user for confirmation.
        fixits = self._response[ 'fixits' ]
        if ( len( fixits ) > 1 or
             ( len( fixits ) == 1 and
               self._command == 'FixIt' and
               fixits[ 0 ].get( 'kind' ) != 'quickfix' ) ):
          fixit_index = vimsupport.SelectFromList(
            "FixIt suggestion(s) available at this location. "
            "Which one would you like to apply?",
            [ fixit[ 'text' ] for fixit in fixits ] )
        chosen_fixit = fixits[ fixit_index ]
        if chosen_fixit[ 'resolve' ]:
          self._request_data.update( { 'fixit': chosen_fixit } )
          response = self.PostDataToHandler( self._request_data,
                                             'resolve_fixit' )
          if response is None:
            return
          fixits = response[ 'fixits' ]
          assert len( fixits ) == 1
          chosen_fixit = fixits[ 0 ]

        vimsupport.ReplaceChunks(
          chosen_fixit[ 'chunks' ],
          silent = self._command == 'Format' )
      except RuntimeError as e:
        vimsupport.PostVimMessage( str( e ) )


  def _HandleBasicResponse( self ):
    vimsupport.PostVimMessage( self._response, warning = False )


  def _HandleMessageResponse( self ):
    vimsupport.PostVimMessage( self._response[ 'message' ], warning = False )


  def _HandleDetailedInfoResponse( self, modifiers ):
    vimsupport.WriteToPreviewWindow( self._response[ 'detailed_info' ],
                                     modifiers )


def SendCommandRequestAsync( arguments,
                             extra_data = None,
                             silent = True,
                             location = None ):
  request = CommandRequest( arguments,
                            extra_data = extra_data,
                            silent = silent,
                            location = location )
  request.Start()
  # Don't block
  return request


def SendCommandRequest( arguments,
                        modifiers,
                        buffer_command = DEFAULT_BUFFER_COMMAND,
                        extra_data = None,
                        skip_post_command_action = False ):
  request = SendCommandRequestAsync( arguments,
                                     extra_data = extra_data,
                                     silent = False )
  # Block here to get the response
  if not skip_post_command_action:
    request.RunPostCommandActionsIfNeeded( modifiers, buffer_command )
  return request.Response()


def GetCommandResponse( arguments, extra_data = None ):
  request = SendCommandRequestAsync( arguments,
                                     extra_data = extra_data,
                                     silent = True )
  # Block here to get the response
  return request.StringResponse()


def GetRawCommandResponse( arguments, silent, location = None ):
  request = SendCommandRequestAsync( arguments,
                                     extra_data = None,
                                     silent = silent,
                                     location = location )
  return request.Response()



# Copyright (C) 2011-2019 ycmd contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

import vim
from ycm import vimsupport
from ycmd import utils
from ycmd.completers.completer import Completer
from ycm.client.base_request import BaseRequest

OMNIFUNC_RETURNED_BAD_VALUE = 'Omnifunc returned bad value to YCM!'
OMNIFUNC_NOT_LIST = ( 'Omnifunc did not return a list or a dict with a "words" '
                     ' list when expected.' )


class OmniCompleter( Completer ):
  def __init__( self, user_options ):
    super( OmniCompleter, self ).__init__( user_options )
    self._omnifunc = None


  def SupportedFiletypes( self ):
    return []


  def ShouldUseCache( self ):
    return bool( self.user_options[ 'cache_omnifunc' ] )


  def ShouldUseNow( self, request_data ):
    self._omnifunc = utils.ToUnicode( vim.eval( '&omnifunc' ) )
    if not self._omnifunc:
      return False
    if self.ShouldUseCache():
      return super( OmniCompleter, self ).ShouldUseNow( request_data )
    return self.ShouldUseNowInner( request_data )


  def ShouldUseNowInner( self, request_data ):
    if request_data[ 'force_semantic' ]:
      return True
    disabled_filetypes = self.user_options[
      'filetype_specific_completion_to_disable' ]
    if not vimsupport.CurrentFiletypesEnabled( disabled_filetypes ):
      return False
    return super( OmniCompleter, self ).ShouldUseNowInner( request_data )


  def ComputeCandidates( self, request_data ):
    if self.ShouldUseCache():
      return super( OmniCompleter, self ).ComputeCandidates( request_data )
    if self.ShouldUseNowInner( request_data ):
      return self.ComputeCandidatesInner( request_data )
    return []


  def ComputeCandidatesInner( self, request_data ):
    if not self._omnifunc:
      return []

    # Calling directly the omnifunc may move the cursor position. This is the
    # case with the default Vim omnifunc for C-family languages
    # (ccomplete#Complete) which calls searchdecl to find a declaration. This
    # function is supposed to move the cursor to the found declaration but it
    # doesn't when called through the omni completion mapping (CTRL-X CTRL-O).
    # So, we restore the cursor position after the omnifunc calls.
    line, column = vimsupport.CurrentLineAndColumn()

    try:
      start_column = vimsupport.GetIntValue( self._omnifunc + '(1,"")' )

      # Vim only stops completion if the value returned by the omnifunc is -3 or
      # -2. In other cases, if the value is negative or greater than the current
      # column, the start column is set to the current column; otherwise, the
      # value is used as the start column.
      if start_column in ( -3, -2 ):
        return []
      if start_column < 0 or start_column > column:
        start_column = column

      # Use the start column calculated by the omnifunc, rather than our own
      # interpretation. This is important for certain languages where our
      # identifier detection is either incorrect or not compatible with the
      # behaviour of the omnifunc. Note: do this before calling the omnifunc
      # because it affects the value returned by 'query'.
      request_data[ 'start_column' ] = start_column + 1

      # Vim internally moves the cursor to the start column before calling again
      # the omnifunc. Some omnifuncs like the one defined by the
      # LanguageClient-neovim plugin depend on this behavior to compute the list
      # of candidates.
      vimsupport.SetCurrentLineAndColumn( line, start_column )

      omnifunc_call = [ self._omnifunc,
                        "(0,'",
                        vimsupport.EscapeForVim( request_data[ 'query' ] ),
                        "')" ]
      items = vim.eval( ''.join( omnifunc_call ) )

      if isinstance( items, dict ) and 'words' in items:
        items = items[ 'words' ]

      if not hasattr( items, '__iter__' ):
        raise TypeError( OMNIFUNC_NOT_LIST )

      # Vim allows each item of the list to be either a string or a dictionary
      # but ycmd only supports lists where items are all strings or all
      # dictionaries. Convert all strings into dictionaries.
      for index, item in enumerate( items ):
        # Set the 'equal' field to 1 to disable Vim filtering.
        if not isinstance( item, dict ):
          items[ index ] = {
            'word': item,
            'equal': 1
          }
        else:
          item[ 'equal' ] = 1

      return items

    except ( TypeError, ValueError, vim.error ) as error:
      vimsupport.PostVimMessage(
        OMNIFUNC_RETURNED_BAD_VALUE + ' ' + str( error ) )
      return []

    finally:
      vimsupport.SetCurrentLineAndColumn( line, column )


  def FilterAndSortCandidatesInner( self, candidates, sort_property, query ):
    request_data = {
      'candidates': candidates,
      'sort_property': sort_property,
      'query': query
    }

    response = BaseRequest().PostDataToHandler( request_data,
                                                'filter_and_sort_candidates' )
    return response if response is not None else []



# Copyright (C) 2020 YouCompleteMe contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

from unittest import TestCase
from unittest.mock import MagicMock
from hamcrest import assert_that, has_entries

from ycm.client.omni_completion_request import OmniCompletionRequest


def BuildOmnicompletionRequest( results, start_column = 1 ):
  omni_completer = MagicMock()
  omni_completer.ComputeCandidates = MagicMock( return_value = results )

  request_data = {
    'line_num': 1,
    'column_num': 1,
    'start_column': start_column
  }
  request = OmniCompletionRequest( omni_completer, request_data )
  request.Start()

  return request


class OmniCompletionRequestTest( TestCase ):
  def test_Done_AlwaysTrue( self ):
    request = BuildOmnicompletionRequest( [] )

    assert_that( request.Done() )


  def test_Response_FromOmniCompleter( self ):
    results = [ { "word": "test" } ]
    request = BuildOmnicompletionRequest( results )

    assert_that( request.Response(), has_entries( {
      'line': 1,
      'column': 1,
      'completion_start_column': 1,
      'completions': results
    } ) )



# Copyright (C) 2016, Davit Samvelyan
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

from ycm import vimsupport
from ycm.client.event_notification import EventNotification
from ycm.diagnostic_interface import DiagnosticInterface
from ycm.semantic_highlighting import SemanticHighlighting
from ycm.inlay_hints import InlayHints


# Emulates Vim buffer
# Used to store buffer related information like diagnostics, latest parse
# request. Stores buffer change tick at the parse request moment, allowing
# to effectively determine whether reparse is needed for the buffer.
class Buffer:

  def __init__( self, bufnr, user_options, filetypes ):
    self._number = bufnr
    self._parse_tick = 0
    self._handled_tick = 0
    self._parse_request = None
    self._should_resend = False
    self._diag_interface = DiagnosticInterface( bufnr, user_options )
    self._open_loclist_on_ycm_diags = user_options[
                                        'open_loclist_on_ycm_diags' ]
    self.semantic_highlighting = SemanticHighlighting( bufnr )
    self.inlay_hints = InlayHints( bufnr )
    self.UpdateFromFileTypes( filetypes )


  def FileParseRequestReady( self, block = False ):
    return ( bool( self._parse_request ) and
             ( block or self._parse_request.Done() ) )


  def SendParseRequest( self, extra_data ):
    # Don't send a parse request if one is in progress
    if self._parse_request is not None and not self._parse_request.Done():
      self._should_resend = True
      return

    self._should_resend = False

    self._parse_request = EventNotification( 'FileReadyToParse',
                                             extra_data = extra_data )
    self._parse_request.Start()
    # Decrement handled tick to ensure correct handling when we are forcing
    # reparse on buffer visit and changed tick remains the same.
    self._handled_tick -= 1
    self._parse_tick = self._ChangedTick()


  def ParseRequestPending( self ):
    return bool( self._parse_request ) and not self._parse_request.Done()


  def NeedsReparse( self ):
    return self._parse_tick != self._ChangedTick()


  def ShouldResendParseRequest( self ):
    return ( self._should_resend
             or ( bool( self._parse_request )
                  and self._parse_request.ShouldResend() ) )


  def UpdateDiagnostics( self, force = False ):
    if force or not self._async_diags:
      self.UpdateWithNewDiagnostics( self._parse_request.Response(), False )
    else:
      # We need to call the response method, because it might throw an exception
      # or require extra config confirmation, even if we don't actually use the
      # diagnostics.
      self._parse_request.Response()


  def UpdateWithNewDiagnostics( self, diagnostics, async_message ):
    self._async_diags = async_message
    self._diag_interface.UpdateWithNewDiagnostics(
        diagnostics,
        not self._async_diags and self._open_loclist_on_ycm_diags )


  def UpdateMatches( self ):
    self._diag_interface.UpdateMatches()


  def PopulateLocationList( self, open_on_edit = False ):
    return self._diag_interface.PopulateLocationList( open_on_edit )


  def GetResponse( self ):
    return self._parse_request.Response()


  def IsResponseHandled( self ):
    return self._handled_tick == self._parse_tick


  def MarkResponseHandled( self ):
    self._handled_tick = self._parse_tick


  def OnCursorMoved( self ):
    self._diag_interface.OnCursorMoved()


  def GetErrorCount( self ):
    return self._diag_interface.GetErrorCount()


  def GetWarningCount( self ):
    return self._diag_interface.GetWarningCount()


  def RefreshDiagnosticsUI( self ):
    return self._diag_interface.RefreshDiagnosticsUI()


  def ClearDiagnosticsUI( self ):
    return self._diag_interface.ClearDiagnosticsUI()


  def DiagnosticsForLine( self, line_number ):
    return self._diag_interface.DiagnosticsForLine( line_number )


  def UpdateFromFileTypes( self, filetypes ):
    self._filetypes = filetypes
    # We will set this to true if we ever receive any diagnostics asyncronously.
    self._async_diags = False


  def _ChangedTick( self ):
    return vimsupport.GetBufferChangedTick( self._number )


class BufferDict( dict ):

  def __init__( self, user_options ):
    self._user_options = user_options


  def __missing__( self, key ):
    # Python does not allow to return assignment operation result directly
    new_value = self[ key ] = Buffer(
      key,
      self._user_options,
      vimsupport.GetBufferFiletypes( key ) )

    return new_value



# Copyright (C) 2017-2018 YouCompleteMe Contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

from ycm.tests.test_utils import MockVimBuffers, MockVimModule, VimBuffer
MockVimModule()

from hamcrest import assert_that, has_entry
from unittest import TestCase
from unittest.mock import patch
from ycm.client.base_request import BuildRequestData


class BaseRequestTest( TestCase ):
  @patch( 'ycm.client.base_request.GetCurrentDirectory',
          return_value = '/some/dir' )
  def test_BuildRequestData_AddWorkingDir( self, *args ):
    current_buffer = VimBuffer( 'foo' )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      assert_that( BuildRequestData(), has_entry( 'working_dir', '/some/dir' ) )


  @patch( 'ycm.client.base_request.GetCurrentDirectory',
          return_value = '/some/dir' )
  def test_BuildRequestData_AddWorkingDirWithFileName( self, *args ):
    current_buffer = VimBuffer( 'foo' )
    with MockVimBuffers( [ current_buffer ], [ current_buffer ] ):
      assert_that( BuildRequestData( current_buffer.number ),
                   has_entry( 'working_dir', '/some/dir' ) )



# Copyright (C) 2023, YouCompleteMe Contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

import abc

from ycm import vimsupport


class ScrollingBufferRange( object ):
  """Abstraction used by inlay hints and semantic tokens to only request visible
  ranges"""

  # FIXME: Send a request per-disjoint range for this buffer rather than the
  # maximal range. then collaate the results when all responses are returned
  def __init__( self, bufnr ):
    self._bufnr = bufnr
    self._tick = -1
    self._request = None
    self._last_requested_range = None


  def Ready( self ):
    return self._request is not None and self._request.Done()


  def Request( self, force=False ):
    if self._request and not self.Ready():
      return True

    # Check to see if the buffer ranges would actually change anything visible.
    # This avoids a round-trip for every single line scroll event
    if ( not force and
         self._tick == vimsupport.GetBufferChangedTick( self._bufnr ) and
         vimsupport.VisibleRangeOfBufferOverlaps(
           self._bufnr,
           self._last_requested_range ) ):
      return False # don't poll

    # FIXME: This call is duplicated in the call to VisibleRangeOfBufferOverlaps
    #  - remove the expansion param
    #  - look up the actual visible range, then call this function
    #  - if not overlapping, do the factor expansion and request
    self._last_requested_range = vimsupport.RangeVisibleInBuffer( self._bufnr )
    # If this is false, either the self._bufnr is not a valid buffer number or
    # the buffer is not visible in any window.
    # Since this is called asynchronously, a user may bwipeout a buffer with
    # self._bufnr number between polls.
    if self._last_requested_range is None:
      return False

    self._tick = vimsupport.GetBufferChangedTick( self._bufnr )

    # We'll never use the last response again, so clear it
    self._latest_response = None
    self._request = self._NewRequest( self._last_requested_range )
    self._request.Start()
    return True


  def Update( self ):
    if not self._request:
      # Nothing to update
      return True

    assert self.Ready()

    # We're ready to use this response. Clear the request (to avoid repeatedly
    # re-polling).
    self._latest_response = self._request.Response()
    self._request = None

    if self._tick != vimsupport.GetBufferChangedTick( self._bufnr ):
      # Buffer has changed, we should ignore the data and retry
      self.Request( force=True )
      return False # poll again

    self._Draw()

    # No need to re-poll
    return True


  def Refresh( self ):
    if self._tick != vimsupport.GetBufferChangedTick( self._bufnr ):
      # stale data
      return

    if self._request is not None:
      # request in progress; we''l handle refreshing when it's done.
      return

    self._Draw()


  def GrowRangeIfNeeded( self, rng ):
    """When processing results, we may receive a wider range than requested. In
    that case, grow our 'last requested' range to minimise requesting more
    frequently than we need to."""
    # Note: references (pointers) so no need to re-assign
    rmin = self._last_requested_range[ 'start' ]
    rmax = self._last_requested_range[ 'end' ]

    start = rng[ 'start' ]
    end = rng[ 'end' ]

    if rmin[ 'line_num' ] is None or start[ 'line_num' ] < rmin[ 'line_num' ]:
      rmin[ 'line_num' ] = start[ 'line_num' ]
      rmin[ 'column_num' ] = start[ 'column_num' ]
    elif start[ 'line_num' ] == rmin[ 'line_num' ]:
      rmin[ 'column_num' ] = min( start[ 'column_num' ],
                                  rmin[ 'column_num' ] )

    if rmax[ 'line_num' ] is None or end[ 'line_num' ] > rmax[ 'line_num' ]:
      rmax[ 'line_num' ] = end[ 'line_num' ]
      rmax[ 'column_num' ] = end[ 'column_num' ]
    elif end[ 'line_num' ] == rmax[ 'line_num' ]:
      rmax[ 'column_num' ] = max( end[ 'column_num' ], rmax[ 'column_num' ] )


  # API; just implement the following, using self._bufnr and
  # self._latest_response as required

  @abc.abstractmethod
  def _NewRequest( self, request_range ):
    # prepare a new request_data and return it
    pass


  @abc.abstractmethod
  def _Draw( self ):
    # actuall paint the properties
    pass



# Copyright (C) 2013 Google Inc.
#               2016 YouCompleteMe contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

from ycm.tests.test_utils import MockVimModule
MockVimModule()

import os
from hamcrest import assert_that, contains_inanyorder, has_item, has_items
from unittest import TestCase
from ycm import syntax_parse
from ycmd.utils import ReadFile


def ContentsOfTestFile( test_file ):
  dir_of_script = os.path.dirname( os.path.abspath( __file__ ) )
  full_path_to_test_file = os.path.join( dir_of_script, 'testdata', test_file )
  return ReadFile( full_path_to_test_file )


class SyntaxTest( TestCase ):
  def test_KeywordsFromSyntaxListOutput_PythonSyntax( self ):
    expected_keywords = ( 'bytearray', 'IndexError', 'all', 'help', 'vars',
        'SyntaxError', 'global', 'elif', 'unicode', 'sorted', 'memoryview',
        'isinstance', 'except', 'nonlocal', 'NameError', 'finally',
        'BytesWarning', 'dict', 'IOError', 'pass', 'oct', 'bin', 'SystemExit',
        'return', 'StandardError', 'format', 'TabError', 'break', 'next',
        'not', 'UnicodeDecodeError', 'False', 'RuntimeWarning', 'list', 'iter',
        'try', 'reload', 'Warning', 'round', 'dir', 'cmp', 'set', 'bytes',
        'UnicodeTranslateError', 'intern', 'issubclass', 'yield', 'Ellipsis',
        'hash', 'locals', 'BufferError', 'slice', 'for', 'FloatingPointError',
        'sum', 'VMSError', 'getattr', 'abs', 'print', 'import', 'True',
        'FutureWarning', 'ImportWarning', 'None', 'EOFError', 'len',
        'frozenset', 'ord', 'super', 'raise', 'TypeError', 'KeyboardInterrupt',
        'UserWarning', 'filter', 'range', 'staticmethod', 'SystemError', 'or',
        'BaseException', 'pow', 'RuntimeError', 'float', 'MemoryError',
        'StopIteration', 'globals', 'divmod', 'enumerate', 'apply',
        'LookupError', 'open', 'basestring', 'from', 'UnicodeError', 'zip',
        'hex', 'long', 'IndentationError', 'int', 'chr', '__import__', 'type',
        'Exception', 'continue', 'tuple', 'reduce', 'reversed', 'else',
        'assert', 'UnicodeEncodeError', 'input', 'with', 'hasattr', 'delattr',
        'setattr', 'raw_input', 'PendingDeprecationWarning', 'compile',
        'ArithmeticError', 'while', 'del', 'str', 'property', 'def', 'and',
        'GeneratorExit', 'ImportError', 'xrange', 'is', 'EnvironmentError',
        'KeyError', 'coerce', 'SyntaxWarning', 'file', 'in', 'unichr', 'ascii',
        'any', 'as', 'if', 'OSError', 'DeprecationWarning', 'min',
        'UnicodeWarning', 'execfile', 'id', 'complex', 'bool', 'ValueError',
        'NotImplemented', 'map', 'exec', 'buffer', 'max', 'class', 'object',
        'repr', 'callable', 'ZeroDivisionError', 'eval', '__debug__',
        'ReferenceError', 'AssertionError', 'classmethod', 'UnboundLocalError',
        'NotImplementedError', 'lambda', 'AttributeError', 'OverflowError',
        'WindowsError' )

    assert_that( syntax_parse._KeywordsFromSyntaxListOutput(
                     ContentsOfTestFile( 'python_syntax' ) ),
                 contains_inanyorder( *expected_keywords ) )


  def test_KeywordsFromSyntaxListOutput_CppSyntax( self ):
    expected_keywords = (
      'int_fast32_t', 'FILE', 'size_t', 'bitor', 'typedef', 'const', 'struct',
      'uint8_t', 'fpos_t', 'thread_local', 'unsigned', 'uint_least16_t', 'do',
      'intptr_t', 'uint_least64_t', 'return', 'auto', 'void', '_Complex',
      'break', '_Alignof', 'not', 'using', '_Static_assert', '_Thread_local',
      'public', 'uint_fast16_t', 'this', 'continue', 'char32_t', 'int16_t',
      'intmax_t', 'static', 'clock_t', 'sizeof', 'int_fast64_t', 'mbstate_t',
      'try', 'xor', 'uint_fast32_t', 'int_least8_t', 'div_t', 'volatile',
      'template', 'char16_t', 'new', 'ldiv_t', 'int_least16_t', 'va_list',
      'uint_least8_t', 'goto', 'noreturn', 'enum', 'static_assert', 'bitand',
      'compl', 'imaginary', 'jmp_buf', 'throw', 'asm', 'ptrdiff_t', 'uint16_t',
      'or', 'uint_fast8_t', '_Bool', 'int32_t', 'float', 'private', 'restrict',
      'wint_t', 'operator', 'not_eq', '_Imaginary', 'alignas', 'union', 'long',
      'uint_least32_t', 'int_least64_t', 'friend', 'uintptr_t', 'int8_t',
      'else', 'export', 'int_fast8_t', 'catch', 'true', 'case', 'default',
      'double', '_Noreturn', 'signed', 'typename', 'while', 'protected',
      'wchar_t', 'wctrans_t', 'uint64_t', 'delete', 'and', 'register', 'false',
      'int', 'uintmax_t', 'off_t', 'char', 'int64_t', 'int_fast16_t', 'DIR',
      '_Atomic', 'time_t', 'xor_eq', 'namespace', 'virtual', 'complex', 'bool',
      'mutable', 'if', 'int_least32_t', 'sig_atomic_t', 'and_eq', 'ssize_t',
      'alignof', '_Alignas', '_Generic', 'extern', 'class', 'typeid', 'short',
      'for', 'uint_fast64_t', 'wctype_t', 'explicit', 'or_eq', 'switch',
      'uint32_t', 'inline' )

    assert_that( syntax_parse._KeywordsFromSyntaxListOutput(
                     ContentsOfTestFile( 'cpp_syntax' ) ),
                 contains_inanyorder( *expected_keywords ) )


  def test_KeywordsFromSyntaxListOutput_JavaSyntax( self ):
    expected_keywords = (
      'code', 'text', 'cols', 'datetime', 'disabled', 'shape', 'codetype',
      'alt', 'compact', 'style', 'valuetype', 'short', 'finally', 'continue',
      'extends', 'valign', 'bordercolor', 'do', 'return', 'rel', 'rules',
      'void', 'nohref', 'abbr', 'background', 'scrolling', 'instanceof',
      'name', 'summary', 'try', 'default', 'noshade', 'coords', 'dir', 'frame',
      'usemap', 'ismap', 'static', 'hspace', 'vlink', 'for', 'selected', 'rev',
      'vspace', 'content', 'method', 'version', 'volatile', 'above', 'new',
      'charoff', 'public', 'alink', 'enum', 'codebase', 'if', 'noresize',
      'interface', 'checked', 'byte', 'super', 'throw', 'src', 'language',
      'package', 'standby', 'script', 'longdesc', 'maxlength', 'cellpadding',
      'throws', 'tabindex', 'color', 'colspan', 'accesskey', 'float', 'while',
      'private', 'height', 'boolean', 'wrap', 'prompt', 'nowrap', 'size',
      'rows', 'span', 'clip', 'bgcolor', 'top', 'long', 'start', 'scope',
      'scheme', 'type', 'final', 'lang', 'visibility', 'else', 'assert',
      'transient', 'link', 'catch', 'true', 'serializable', 'target', 'lowsrc',
      'this', 'double', 'align', 'value', 'cite', 'headers', 'below',
      'protected', 'declare', 'classid', 'defer', 'false', 'synchronized',
      'int', 'abstract', 'accept', 'hreflang', 'char', 'border', 'id',
      'native', 'rowspan', 'charset', 'archive', 'strictfp', 'readonly',
      'axis', 'cellspacing', 'profile', 'multiple', 'object', 'action',
      'pagex', 'pagey', 'marginheight', 'data', 'class', 'frameborder',
      'enctype', 'implements', 'break', 'gutter', 'url', 'clear', 'face',
      'switch', 'marginwidth', 'width', 'left' )

    assert_that( syntax_parse._KeywordsFromSyntaxListOutput(
                     ContentsOfTestFile( 'java_syntax' ) ),
                 contains_inanyorder( *expected_keywords ) )


  def test_KeywordsFromSyntaxListOutput_PhpSyntax_ContainsFunctions( self ):
    assert_that( syntax_parse._KeywordsFromSyntaxListOutput(
                     ContentsOfTestFile( 'php_syntax' ) ),
                 has_items( 'array_change_key_case' ) )


  def test_KeywordsFromSyntaxListOutput_PhpSyntax_ContainsPreProc( self ):
    assert_that( syntax_parse._KeywordsFromSyntaxListOutput(
                     ContentsOfTestFile( 'php_syntax' ) ),
                 has_items( 'skip', 'function' ) )


  def test_KeywordsFromSyntaxListOutput_Basic( self ):
    assert_that( syntax_parse._KeywordsFromSyntaxListOutput( """
foogroup xxx foo bar
             zoo goo
             links to Statement""" ),
                contains_inanyorder( 'foo', 'bar', 'zoo', 'goo' ) )


  def test_KeywordsFromSyntaxListOutput_Function( self ):
    assert_that( syntax_parse._KeywordsFromSyntaxListOutput( """
foogroup xxx foo bar
             zoo goo
             links to Function""" ),
                 contains_inanyorder( 'foo', 'bar', 'zoo', 'goo' ) )


  def test_KeywordsFromSyntaxListOutput_ContainedArgAllowed( self ):
    assert_that( syntax_parse._KeywordsFromSyntaxListOutput( """
phpFunctions   xxx contained gzclose yaz_syntax html_entity_decode fbsql_read_blob png2wbmp mssql_init cpdf_set_title gztell fbsql_insert_id empty cpdf_restore mysql_field_type closelog swftext ldap_search curl_errno gmp_div_r mssql_data_seek getmyinode printer_draw_pie mcve_initconn ncurses_getmaxyx defined
                   contained replace_child has_attributes specified insertdocument assign node_name hwstat addshape get_attribute_node html_dump_mem userlist
                   links to Function""" ), # noqa
                has_items( 'gzclose', 'userlist', 'ldap_search' ) )


  def test_KeywordsFromSyntaxListOutput_JunkIgnored( self ):
    assert_that( syntax_parse._KeywordsFromSyntaxListOutput( """
--- Syntax items ---
foogroup xxx foo bar
             zoo goo
             links to Statement
Spell        cluster=NONE
NoSpell      cluster=NONE""" ),
                 contains_inanyorder( 'foo', 'bar', 'zoo', 'goo' ) )


  def test_KeywordsFromSyntaxListOutput_MultipleStatementGroups( self ):
    assert_that( syntax_parse._KeywordsFromSyntaxListOutput( """
foogroup xxx foo bar
             links to Statement
bargroup xxx zoo goo
             links to Statement""" ),
                 contains_inanyorder( 'foo', 'bar', 'zoo', 'goo' ) )


  def test_KeywordsFromSyntaxListOutput_StatementAndTypeGroups( self ):
    assert_that( syntax_parse._KeywordsFromSyntaxListOutput( """
foogroup xxx foo bar
             links to Statement
bargroup xxx zoo goo
             links to Type""" ),
                 contains_inanyorder( 'foo', 'bar', 'zoo', 'goo' ) )


  def test_KeywordsFromSyntaxListOutput_StatementHierarchy( self ):
    assert_that( syntax_parse._KeywordsFromSyntaxListOutput( """
baa xxx foo bar
        links to Foo
Foo xxx zoo goo
        links to Bar
Bar xxx qux moo
        links to Statement""" ),
                 contains_inanyorder( 'foo', 'bar', 'zoo',
                                      'goo', 'qux', 'moo' ) )


  def test_KeywordsFromSyntaxListOutput_TypeHierarchy( self ):
    assert_that( syntax_parse._KeywordsFromSyntaxListOutput( """
baa xxx foo bar
        links to Foo
Foo xxx zoo goo
        links to Bar
Bar xxx qux moo
        links to Type""" ),
                contains_inanyorder( 'foo', 'bar', 'zoo',
                                     'goo', 'qux', 'moo' ) )


  def test_KeywordsFromSyntaxListOutput_StatementAndTypeHierarchy( self ):
    assert_that( syntax_parse._KeywordsFromSyntaxListOutput( """
tBaa xxx foo bar
         links to tFoo
tFoo xxx zoo goo
         links to tBar
tBar xxx qux moo
         links to Type
sBaa xxx na bar
         links to sFoo
sFoo xxx zoo nb
         links to sBar
sBar xxx qux nc
         links to Statement""" ),
                contains_inanyorder( 'foo', 'bar', 'zoo', 'goo', 'qux', 'moo',
                                     'na', 'nb', 'nc' ) )


  def test_SyntaxGroupsFromOutput_Basic( self ):
    assert_that( syntax_parse._SyntaxGroupsFromOutput( """
foogroup xxx foo bar
             zoo goo
             links to Statement""" ),
                 has_item( 'foogroup' ) )


  def test_ExtractKeywordsFromGroup_Basic( self ):
    assert_that( syntax_parse._ExtractKeywordsFromGroup(
                   syntax_parse.SyntaxGroup( '', [
                     'foo bar',
                     'zoo goo',
                   ] ) ),
                 contains_inanyorder( 'foo', 'bar', 'zoo', 'goo' ) )


  def test_ExtractKeywordsFromGroup_Commas( self ):
    assert_that( syntax_parse._ExtractKeywordsFromGroup(
                   syntax_parse.SyntaxGroup( '', [
                     'foo, bar,',
                     'zoo goo',
                   ] ) ),
                 contains_inanyorder( 'foo', 'bar', 'zoo', 'goo' ) )


  def test_ExtractKeywordsFromGroup_WithLinksTo( self ):
    assert_that( syntax_parse._ExtractKeywordsFromGroup(
                   syntax_parse.SyntaxGroup( '', [
                     'foo bar',
                     'zoo goo',
                     'links to Statement'
                   ] ) ),
                 contains_inanyorder( 'foo', 'bar', 'zoo', 'goo' ) )


  def test_ExtractKeywordsFromGroup_KeywordStarts( self ):
    assert_that( syntax_parse._ExtractKeywordsFromGroup(
                   syntax_parse.SyntaxGroup( '', [
                     'foo bar',
                     'contained boo baa',
                     'zoo goo',
                   ] ) ),
                 contains_inanyorder( 'foo', 'bar', 'boo',
                                      'baa', 'zoo', 'goo' ) )


  def test_ExtractKeywordsFromGroup_KeywordMiddle( self ):
    assert_that( syntax_parse._ExtractKeywordsFromGroup(
                   syntax_parse.SyntaxGroup( '', [
                     'foo contained bar',
                     'zoo goo'
                   ] ) ),
                 contains_inanyorder( 'foo', 'contained',
                                      'bar', 'zoo', 'goo' ) )


  def test_ExtractKeywordsFromGroup_KeywordAssign( self ):
    assert_that( syntax_parse._ExtractKeywordsFromGroup(
                   syntax_parse.SyntaxGroup( '', [
                     'nextgroup=zoo skipwhite foo bar',
                     'zoo goo',
                   ] ) ),
                 contains_inanyorder( 'foo', 'bar', 'zoo', 'goo' ) )


  def test_ExtractKeywordsFromGroup_KeywordAssignAndMiddle( self ):
    assert_that( syntax_parse._ExtractKeywordsFromGroup(
                   syntax_parse.SyntaxGroup( '', [
                     'nextgroup=zoo foo skipnl bar',
                     'zoo goo',
                   ] ) ),
                 contains_inanyorder( 'foo', 'skipnl', 'bar', 'zoo', 'goo' ) )


  def test_ExtractKeywordsFromGroup_KeywordWithoutNextgroup( self ):
    assert_that( syntax_parse._ExtractKeywordsFromGroup(
                   syntax_parse.SyntaxGroup( '', [
                     'skipempty foo bar',
                     'zoo goo',
                   ] ) ),
                 contains_inanyorder( 'skipempty', 'foo',
                                      'bar', 'zoo', 'goo' ) )


  def test_ExtractKeywordsFromGroup_ContainedSyntaxArgAllowed( self ):
    assert_that( syntax_parse._ExtractKeywordsFromGroup(
                   syntax_parse.SyntaxGroup( '', [
                     'contained foo zoq',
                     'contained bar goo',
                     'far'
                   ] ) ),
                 contains_inanyorder( 'foo', 'zoq', 'bar', 'goo', 'far' ) )



# Copyright 2009 Brian Quinlan. All Rights Reserved.
#   Licensed to PSF under a Contributor Agreement.
#
# Copyright (C) 2013  Google Inc.
#   Changes to this file are licensed under the same terms as the original file
#   (the Python Software Foundation License).


import threading
import weakref
import sys

from concurrent.futures import _base

try:
  import queue
except ImportError:
  import Queue as queue


# This file provides an UnsafeThreadPoolExecutor, which operates exactly like
# the upstream Python version of ThreadPoolExecutor with one exception: it
# doesn't wait for worker threads to finish before shutting down the Python
# interpreter.
#
# This is dangerous for many workloads, but fine for some (like when threads
# only send network requests). The YCM workload is one of those workloads where
# it's safe (the aforementioned network requests case).

class _WorkItem:
  def __init__( self, future, fn, args, kwargs ):
    self.future = future
    self.fn = fn
    self.args = args
    self.kwargs = kwargs

  def run( self ):
    if not self.future.set_running_or_notify_cancel():
      return

    try:
      result = self.fn( *self.args, **self.kwargs )
    except BaseException:
      e = sys.exc_info()[ 1 ]
      self.future.set_exception( e )
    else:
      self.future.set_result( result )


def _worker( executor_reference, work_queue ):
  try:
    while True:
      work_item = work_queue.get( block=True )
      if work_item is not None:
        work_item.run()
        continue
      executor = executor_reference()
      # Exit if:
      #   - The executor that owns the worker has been collected OR
      #   - The executor that owns the worker has been shutdown.
      if executor is None or executor._shutdown:
        # Notice other workers
        work_queue.put( None )
        return
      del executor
  except BaseException:
    _base.LOGGER.critical( 'Exception in worker', exc_info=True )


class UnsafeThreadPoolExecutor( _base.Executor ):
  def __init__( self, max_workers ):
    """Initializes a new ThreadPoolExecutor instance.

    Args:
        max_workers: The maximum number of threads that can be used to
            execute the given calls.
    """
    self._max_workers = max_workers
    self._work_queue = queue.Queue()
    self._threads = set()
    self._shutdown = False
    self._shutdown_lock = threading.Lock()

  def submit( self, fn, *args, **kwargs ):
    with self._shutdown_lock:
      if self._shutdown:
        raise RuntimeError( 'cannot schedule new futures after shutdown' )

      f = _base.Future()
      w = _WorkItem( f, fn, args, kwargs )

      self._work_queue.put( w )
      self._adjust_thread_count()
      return f
  submit.__doc__ = _base.Executor.submit.__doc__

  def _adjust_thread_count( self ):
    # When the executor gets lost, the weakref callback will wake up
    # the worker threads.
    def weakref_cb( _, q=self._work_queue ):
      q.put( None )
    # TODO(bquinlan): Should avoid creating new threads if there are more
    # idle threads than items in the work queue.
    if len( self._threads ) < self._max_workers:
      t = threading.Thread( target=_worker,
                            args=( weakref.ref( self, weakref_cb ),
                                   self._work_queue ) )
      t.daemon = True
      t.start()
      self._threads.add( t )

  def shutdown( self, wait=True ):
    with self._shutdown_lock:
      self._shutdown = True
      self._work_queue.put( None )
    if wait:
      for t in self._threads:
        t.join()
  shutdown.__doc__ = _base.Executor.shutdown.__doc__



# Copyright (C) 2020, YouCompleteMe Contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.


from ycm.client.semantic_tokens_request import SemanticTokensRequest
from ycm.client.base_request import BuildRequestData
from ycm import vimsupport
from ycm import text_properties as tp
from ycm import scrolling_range as sr

import vim


HIGHLIGHT_GROUP = {
  'namespace': 'Type',
  'type': 'Type',
  'class': 'Structure',
  'enum': 'Structure',
  'interface': 'Structure',
  'struct': 'Structure',
  'typeParameter': 'Identifier',
  'parameter': 'Identifier',
  'variable': 'Identifier',
  'property': 'Identifier',
  'enumMember': 'Identifier',
  'enumConstant': 'Constant',
  'event': 'Identifier',
  'function': 'Function',
  'member': 'Identifier',
  'macro': 'Macro',
  'method': 'Function',
  'keyword': 'Keyword',
  'modifier': 'Keyword',
  'comment': 'Comment',
  'string': 'String',
  'number': 'Number',
  'regexp': 'String',
  'operator': 'Operator',
  'decorator': 'Special',
  'unknown': 'Normal',

  # These are not part of the spec, but are used by clangd
  'bracket': 'Normal',
  'concept': 'Type',
  # These are not part of the spec, but are used by jdt.ls
  'annotation': 'Macro',
}
REPORTED_MISSING_TYPES = set()


def Initialise():
  if vimsupport.VimIsNeovim():
    return

  props = tp.GetTextPropertyTypes()
  if 'YCM_HL_UNKNOWN' not in props:
    tp.AddTextPropertyType( 'YCM_HL_UNKNOWN',
                            highlight = 'WarningMsg',
                            priority = 0 )

  for token_type, group in HIGHLIGHT_GROUP.items():
    prop = f'YCM_HL_{ token_type }'
    if prop not in props and vimsupport.GetIntValue(
        f"hlexists( '{ vimsupport.EscapeForVim( group ) }' )" ):
      tp.AddTextPropertyType( prop,
                              highlight = group,
                              priority = 0 )


# "arbitrary" base id
NEXT_TEXT_PROP_ID = 70784


def NextPropID():
  global NEXT_TEXT_PROP_ID
  try:
    return NEXT_TEXT_PROP_ID
  finally:
    NEXT_TEXT_PROP_ID += 1



class SemanticHighlighting( sr.ScrollingBufferRange ):
  """Stores the semantic highlighting state for a Vim buffer"""

  def __init__( self, bufnr ):
    self._prop_id = NextPropID()
    super().__init__( bufnr )


  def _NewRequest( self, request_range ):
    request: dict = BuildRequestData( self._bufnr )
    request[ 'range' ] = request_range
    return SemanticTokensRequest( request )


  def _Draw( self ):
    # We requested a snapshot
    tokens = self._latest_response.get( 'tokens', [] )

    prev_prop_id = self._prop_id
    self._prop_id = NextPropID()

    for token in tokens:
      prop_type = f"YCM_HL_{ token[ 'type' ] }"
      rng = token[ 'range' ]
      self.GrowRangeIfNeeded( rng )

      try:
        tp.AddTextProperty( self._bufnr, self._prop_id, prop_type, rng )
      except vim.error as e:
        if 'E971:' in str( e ): # Text property doesn't exist
          if token[ 'type' ] not in REPORTED_MISSING_TYPES:
            REPORTED_MISSING_TYPES.add( token[ 'type' ] )
            vimsupport.PostVimMessage(
              f"Token type { token[ 'type' ] } not supported. "
              f"Define property type { prop_type }. "
              f"See :help youcompleteme-customising-highlight-groups" )
        else:
          raise e

    tp.ClearTextProperties( self._bufnr, prop_id = prev_prop_id )



# Copyright (C) 2017 YouCompleteMe Contributors
#
# This file is part of YouCompleteMe.
#
# YouCompleteMe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# YouCompleteMe is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with YouCompleteMe.  If not, see <http://www.gnu.org/licenses/>.

from ycm.tests.test_utils import MockVimModule
MockVimModule()

from hamcrest import assert_that, equal_to
from unittest import TestCase
from unittest.mock import patch, call

from ycm.client.messages_request import _HandlePollResponse
from ycm.tests.test_utils import ExtendedMock


class MessagesRequestTest( TestCase ):
  def test_HandlePollResponse_NoMessages( self ):
    assert_that( _HandlePollResponse( True, None ), equal_to( True ) )

    # Other non-False responses mean the same thing
    assert_that( _HandlePollResponse( '', None ), equal_to( True ) )
    assert_that( _HandlePollResponse( 1, None ), equal_to( True ) )
    assert_that( _HandlePollResponse( {}, None ), equal_to( True ) )


  def test_HandlePollResponse_PollingNotSupported( self ):
    assert_that( _HandlePollResponse( False, None ), equal_to( False ) )

    # 0 is not False
    assert_that( _HandlePollResponse( 0, None ), equal_to( True ) )


  @patch( 'ycm.client.messages_request.PostVimMessage',
          new_callable = ExtendedMock )
  def test_HandlePollResponse_SingleMessage( self, post_vim_message ):
    assert_that( _HandlePollResponse( [ { 'message': 'this is a message' } ] ,
                                      None ),
                 equal_to( True ) )

    post_vim_message.assert_has_exact_calls( [
      call( 'this is a message', warning=False, truncate=True )
    ] )


  @patch( 'ycm.client.messages_request.PostVimMessage',
          new_callable = ExtendedMock )
  def test_HandlePollResponse_MultipleMessages( self, post_vim_message ):
    assert_that( _HandlePollResponse( [ { 'message': 'this is a message' },
                                        { 'message': 'this is another one' } ] ,
                                      None ),
                 equal_to( True ) )

    post_vim_message.assert_has_exact_calls( [
      call( 'this is a message', warning=False, truncate=True ),
      call( 'this is another one', warning=False, truncate=True )
    ] )


  def test_HandlePollResponse_SingleDiagnostic( self ):
    diagnostics_handler = ExtendedMock()
    messages = [
      { 'filepath': 'foo', 'diagnostics': [ 'PLACEHOLDER' ] },
    ]
    assert_that( _HandlePollResponse( messages, diagnostics_handler ),
                 equal_to( True ) )
    diagnostics_handler.UpdateWithNewDiagnosticsForFile.assert_has_exact_calls(
      [
        call( 'foo', [ 'PLACEHOLDER' ] )
      ] )


  def test_HandlePollResponse_MultipleDiagnostics( self ):
    diagnostics_handler = ExtendedMock()
    messages = [
      { 'filepath': 'foo', 'diagnostics': [ 'PLACEHOLDER1' ] },
      { 'filepath': 'bar', 'diagnostics': [ 'PLACEHOLDER2' ] },
      { 'filepath': 'baz', 'diagnostics': [ 'PLACEHOLDER3' ] },
      { 'filepath': 'foo', 'diagnostics': [ 'PLACEHOLDER4' ] },
    ]
    assert_that( _HandlePollResponse( messages, diagnostics_handler ),
                 equal_to( True ) )
    diagnostics_handler.UpdateWithNewDiagnosticsForFile.assert_has_exact_calls(
      [
        call( 'foo', [ 'PLACEHOLDER1' ] ),
        call( 'bar', [ 'PLACEHOLDER2' ] ),
        call( 'baz', [ 'PLACEHOLDER3' ] ),
        call( 'foo', [ 'PLACEHOLDER4' ] )
      ] )


  @patch( 'ycm.client.messages_request.PostVimMessage',
          new_callable = ExtendedMock )
  def test_HandlePollResponse_MultipleMessagesAndDiagnostics(
      self, post_vim_message ):
    diagnostics_handler = ExtendedMock()
    messages = [
      { 'filepath': 'foo', 'diagnostics': [ 'PLACEHOLDER1' ] },
      { 'message': 'On the first day of Christmas, my VimScript gave to me' },
      { 'filepath': 'bar', 'diagnostics': [ 'PLACEHOLDER2' ] },
      { 'message': 'A test file in a Command-T' },
      { 'filepath': 'baz', 'diagnostics': [ 'PLACEHOLDER3' ] },
      { 'message': 'On the second day of Christmas, my VimScript gave to me' },
      { 'filepath': 'foo', 'diagnostics': [ 'PLACEHOLDER4' ] },
      { 'message': 'Two popup menus, and a test file in a Command-T' },
    ]
    assert_that( _HandlePollResponse( messages, diagnostics_handler ),
                 equal_to( True ) )
    diagnostics_handler.UpdateWithNewDiagnosticsForFile.assert_has_exact_calls(
      [
        call( 'foo', [ 'PLACEHOLDER1' ] ),
        call( 'bar', [ 'PLACEHOLDER2' ] ),
        call( 'baz', [ 'PLACEHOLDER3' ] ),
        call( 'foo', [ 'PLACEHOLDER4' ] )
      ] )

    post_vim_message.assert_has_exact_calls( [
      call( 'On the first day of Christmas, my VimScript gave to me',
            warning=False,
            truncate=True ),
      call( 'A test file in a Command-T', warning=False, truncate=True ),
      call( 'On the second day of Christmas, my VimScript gave to me',
            warning=False,
            truncate=True ),
      call( 'Two popup menus, and a test file in a Command-T',
            warning=False,
            truncate=True ),
    ] )
